from abc import ABC, abstractmethod
from json import dumps, load
from pathlib import Path
from typing import Literal, cast

import httpx
from eth_account.messages import encode_defunct
from eth_typing import ABI, AnyAddress, ChecksumAddress, HexStr
from web3 import Web3
from web3.constants import MAX_INT
from web3.eth import Contract
from web3.exceptions import ContractCustomError, TimeExhausted
from web3.middleware import (
    ExtraDataToPOAMiddleware,
    Middleware,
    SignAndSendRawMiddlewareBuilder,
)
from web3.types import TxParams, Wei

from ..types.common import EthAddress, Keccak256
from ..types.web3_types import TransactionReceipt
from ..utilities.config import get_contract_config
from ..utilities.constants import ADDRESS_ZERO, HASH_ZERO, POLYGON
from ..utilities.exceptions import SafeAlreadyDeployedError
from ..utilities.headers import create_relayer_headers
from ..utilities.signing.signer import Signer
from ..utilities.web3.abis.custom_contract_errors import CUSTOM_ERROR_DICT
from ..utilities.web3.helpers import (
    SafeTxn,
    create_proxy_struct,
    create_safe_create_signature,
    get_index_set,
    get_packed_signature,
    get_signature_type_from_runtime_code,
    sign_safe_transaction,
    split_signature,
)


def _load_abi(contract_name: str) -> ABI:
    abi_path = (
        Path(__file__).parent.parent
        / "utilities"
        / "web3"
        / "abis"
        / f"{contract_name}.json"
    )
    with Path.open(abi_path) as f:
        return cast("ABI", load(f))


class BaseWeb3Client(ABC):
    """
    Abstract base class for Polymarket Web3 clients.

    Contains all shared logic for contract interactions, encoding,
    and read operations. Subclasses implement the execution strategy.
    """

    def __init__(
        self,
        private_key: HexStr,
        signature_type: Literal[0, 1, 2],
        chain_id: Literal[137, 80002] = POLYGON,
        rpc_url: str = "https://tenderly.rpc.polygon.community",
    ):
        self.client = httpx.Client(http2=True, timeout=30.0)
        self.w3 = Web3(Web3.HTTPProvider(rpc_url))
        self.w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
        self.w3.middleware_onion.inject(
            cast("Middleware", SignAndSendRawMiddlewareBuilder.build(private_key)),
            layer=0,
        )

        self.account = self.w3.eth.account.from_key(private_key)
        self.signature_type = signature_type

        self.config = get_contract_config(chain_id, neg_risk=False)
        self.neg_risk_config = get_contract_config(chain_id, neg_risk=True)
        self.chain_id = chain_id
        self._setup_contracts()
        self._setup_address()

    def _setup_contracts(self) -> None:
        """Initialize all contract instances."""
        self.usdc_address = Web3.to_checksum_address(self.config.collateral)
        self.usdc_abi = _load_abi("UChildERC20Proxy")
        self.usdc = self._contract(self.usdc_address, self.usdc_abi)

        self.conditional_tokens_address = Web3.to_checksum_address(
            self.config.conditional_tokens
        )
        self.conditional_tokens_abi = _load_abi("ConditionalTokens")
        self.conditional_tokens = self._contract(
            self.conditional_tokens_address, self.conditional_tokens_abi
        )

        self.exchange_address = Web3.to_checksum_address(self.config.exchange)
        self.exchange_abi = _load_abi("CTFExchange")
        self.exchange = self._contract(self.exchange_address, self.exchange_abi)

        self.neg_risk_exchange_address = Web3.to_checksum_address(
            self.neg_risk_config.exchange
        )
        self.neg_risk_exchange_abi = _load_abi("NegRiskCtfExchange")
        self.neg_risk_exchange = self._contract(
            self.neg_risk_exchange_address, self.neg_risk_exchange_abi
        )

        self.neg_risk_adapter_address = Web3.to_checksum_address(
            "0xd91E80cF2E7be2e162c6513ceD06f1dD0dA35296"
        )
        self.neg_risk_adapter_abi = _load_abi("NegRiskAdapter")
        self.neg_risk_adapter = self._contract(
            self.neg_risk_adapter_address, self.neg_risk_adapter_abi
        )

        self.proxy_factory_address = Web3.to_checksum_address(
            "0xaB45c5A4B0c941a2F231C04C3f49182e1A254052"
        )
        self.proxy_factory_abi = _load_abi("ProxyWalletFactory")
        self.proxy_factory = self._contract(
            self.proxy_factory_address, self.proxy_factory_abi
        )

        self.safe_proxy_factory_address = Web3.to_checksum_address(
            "0xaacFeEa03eb1561C4e67d661e40682Bd20E3541b"
        )
        self.safe_proxy_factory_abi = _load_abi("SafeProxyFactory")
        self.safe_proxy_factory = self._contract(
            self.safe_proxy_factory_address, self.safe_proxy_factory_abi
        )

    def _setup_address(self) -> None:
        """Setup address based on signature type."""
        match self.signature_type:
            case 0:
                self.address = self.account.address
            case 1:
                self.address = self.get_poly_proxy_address()
            case 2:
                self.address = self.get_safe_proxy_address()
                self.safe_abi = _load_abi("Safe")
                self.safe = self._contract(self.address, self.safe_abi)

    def _contract(self, address: AnyAddress | str | bytes, abi: ABI) -> Contract:
        """Create contract instance."""
        return self.w3.eth.contract(
            address=Web3.to_checksum_address(address),
            abi=abi,
        )

    def _encode_usdc_approve(self, address: ChecksumAddress) -> str:
        """Encode USDC approval transaction."""
        abi = self.usdc.encode_abi(
            abi_element_identifier="approve",
            args=[address, int(MAX_INT, base=16)],
        )
        return cast("str", abi)

    def _encode_condition_tokens_approve(self, address: ChecksumAddress) -> str:
        """Encode conditional tokens approval transaction."""
        abi = self.conditional_tokens.encode_abi(
            abi_element_identifier="setApprovalForAll",
            args=[address, True],
        )
        return cast("str", abi)

    def _encode_transfer_usdc(self, address: ChecksumAddress, amount: int) -> str:
        """Encode USDC transfer transaction."""
        abi = self.usdc.encode_abi(
            abi_element_identifier="transfer",
            args=[address, amount],
        )
        return cast("str", abi)

    def _encode_transfer_token(
        self, token_id: str, address: ChecksumAddress, amount: int
    ) -> str:
        """Encode token transfer transaction."""
        abi = self.conditional_tokens.encode_abi(
            abi_element_identifier="safeTransferFrom",
            args=[self.address, address, int(token_id), amount, HASH_ZERO],
        )
        return cast("str", abi)

    def _encode_split(self, condition_id: Keccak256, amount: int) -> str:
        """Encode split position transaction."""
        abi = self.conditional_tokens.encode_abi(
            abi_element_identifier="splitPosition",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2], amount],
        )
        return cast("str", abi)

    def _encode_merge(self, condition_id: Keccak256, amount: int) -> str:
        """Encode merge positions transaction."""
        abi = self.conditional_tokens.encode_abi(
            abi_element_identifier="mergePositions",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2], amount],
        )
        return cast("str", abi)

    def _encode_redeem(self, condition_id: Keccak256) -> str:
        """Encode redeem positions transaction."""
        abi = self.conditional_tokens.encode_abi(
            abi_element_identifier="redeemPositions",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2]],
        )
        return cast("str", abi)

    def _encode_redeem_neg_risk(
        self, condition_id: Keccak256, amounts: list[int]
    ) -> str:
        """Encode redeem positions transaction for neg risk."""
        abi = self.neg_risk_adapter.encode_abi(
            abi_element_identifier="redeemPositions",
            args=[condition_id, amounts],
        )
        return cast("str", abi)

    def _encode_convert(
        self, neg_risk_market_id: Keccak256, index_set: int, amount: int
    ) -> str:
        """Encode convert positions transaction."""
        abi = self.neg_risk_adapter.encode_abi(
            abi_element_identifier="convertPositions",
            args=[neg_risk_market_id, index_set, amount],
        )
        return cast("str", abi)

    def _encode_proxy(self, proxy_txn: dict[str, object]) -> str:
        """Encode proxy transaction."""
        abi = self.proxy_factory.encode_abi(
            abi_element_identifier="proxy",
            args=[[proxy_txn]],
        )
        return cast("str", abi)

    def get_base_address(self) -> EthAddress:
        """Get the base EOA address."""
        return cast("EthAddress", self.account.address)

    def get_poly_proxy_address(self, address: EthAddress | None = None) -> EthAddress:
        """Get the Polymarket proxy address."""
        address = address or self.account.address
        result = self.exchange.functions.getPolyProxyWalletAddress(address).call()
        return cast("EthAddress", result)

    def get_safe_proxy_address(self, address: EthAddress | None = None) -> EthAddress:
        """Get the Safe proxy address."""
        address = address or self.account.address
        result = self.safe_proxy_factory.functions.computeProxyAddress(address).call()
        return cast("EthAddress", result)

    def detect_wallet_signature_type(
        self, address: EthAddress
    ) -> Literal[0, 1, 2] | None:
        """
        Detect wallet signature type from an address.

        Returns:
            - 0 for EOA
            - 1 for Polymarket proxy wallet
            - 2 for Safe/Gnosis proxy wallet
            - None for other smart contracts / unknown wallet implementations

        """
        code = self.w3.eth.get_code(self.w3.to_checksum_address(address)).hex().removeprefix("0x").lower()
        return get_signature_type_from_runtime_code(code)

    def get_pol_balance(self) -> float:
        """Get POL balance for the base address associated with the private key."""
        return round(self.w3.eth.get_balance(self.account.address) / 10**18, 4)

    def get_usdc_balance(self, address: EthAddress | None = None) -> float:
        """
        Get USDC balance of an address.

        If no address is given, returns the balance of the instantiated client.
        """
        if address is None:
            address = self.address
        balance_res = self.usdc.functions.balanceOf(address).call()
        return float(balance_res / 1e6)

    def get_token_balance(
        self, token_id: str, address: EthAddress | None = None
    ) -> float:
        """Get token balance of an address."""
        if not address:
            address = self.address
        balance_res = self.conditional_tokens.functions.balanceOf(
            address, int(token_id)
        ).call()
        return float(balance_res / 1e6)

    def get_token_complement(self, token_id: str) -> str | None:
        """Get the complement token ID."""
        try:
            return str(
                self.neg_risk_exchange.functions.getComplement(int(token_id)).call()
            )
        except ContractCustomError as e:
            if e.args[0] in CUSTOM_ERROR_DICT:
                try:
                    return str(
                        self.exchange.functions.getComplement(int(token_id)).call()
                    )
                except ContractCustomError as e2:
                    if e2.args[0] in CUSTOM_ERROR_DICT:
                        msg = f"{CUSTOM_ERROR_DICT[e2.args[0]]}"
                        raise ContractCustomError(msg) from e2
                    return None
            return None

    def get_condition_id_neg_risk(self, question_id: Keccak256) -> Keccak256:
        """
        Get condition ID for a neg risk market.

        Returns a keccak256 hash of the oracle and question id.
        """
        keccak = (
            "0x"
            + self.neg_risk_adapter.functions.getConditionId(question_id).call().hex()
        )

        return cast("Keccak256", keccak)

    @abstractmethod
    def _execute(
        self,
        to: ChecksumAddress,
        data: str,
        operation_name: str,
        metadata: str | None = None,
    ) -> TransactionReceipt:
        """
        Execute a transaction (abstract method).

        Subclasses must implement this to define how transactions are executed
        (on-chain with gas vs gasless via relay).

        Args:
            to: Contract address to call
            data: Encoded transaction data
            operation_name: Name of operation for logging
            metadata: Optional metadata for gasless transactions

        Returns:
            TransactionReceipt

        """

    def split_position(
        self, condition_id: Keccak256, amount: float, neg_risk: bool = True
    ) -> TransactionReceipt:
        """Split USDC into two complementary positions."""
        amount_int = int(amount * 1e6)

        to = (
            self.neg_risk_adapter_address
            if neg_risk
            else self.conditional_tokens_address
        )
        data = self._encode_split(condition_id, amount_int)

        return self._execute(to, data, "Split Position", metadata="split")

    def merge_position(
        self, condition_id: Keccak256, amount: float, neg_risk: bool = True
    ) -> TransactionReceipt:
        """Merge two complementary positions into USDC."""
        amount_int = int(amount * 1e6)

        to = (
            self.neg_risk_adapter_address
            if neg_risk
            else self.conditional_tokens_address
        )
        data = self._encode_merge(condition_id, amount_int)

        return self._execute(to, data, "Merge Position", metadata="merge")

    def redeem_position(
        self, condition_id: Keccak256, amounts: list[float], neg_risk: bool = True
    ) -> TransactionReceipt:
        """
        Redeem positions into USDC.

        Args:
            condition_id: Condition ID
            amounts: List of amounts [x, y] where x is shares of first outcome,
                     y is shares of second outcome
            neg_risk: Whether this is a neg risk market

        """
        int_amounts = [int(amount * 1e6) for amount in amounts]

        to = (
            self.neg_risk_adapter_address
            if neg_risk
            else self.conditional_tokens_address
        )
        data = (
            self._encode_redeem_neg_risk(condition_id, int_amounts)
            if neg_risk
            else self._encode_redeem(condition_id)
        )

        return self._execute(to, data, "Redeem Position", metadata="redeem")

    def convert_positions(
        self,
        question_ids: list[Keccak256],
        amount: float,
    ) -> TransactionReceipt:
        """
        Convert neg risk No positions to Yes positions and USDC.

        Args:
            question_ids: Array of question_ids representing positions to convert
            amount: Number of shares to convert

        """
        amount_int = int(amount * 1e6)
        neg_risk_market_id = question_ids[0][:-2] + "00"

        to = self.neg_risk_adapter_address
        data = self._encode_convert(
            neg_risk_market_id, get_index_set(question_ids), amount_int
        )

        return self._execute(to, data, "Convert Positions", metadata="convert")


class PolymarketWeb3Client(BaseWeb3Client):
    """
    Polymarket Web3 client for on-chain transactions (pays gas).

    Supports:
    - EOA wallets (signature_type=0)
    - Poly proxy wallets (signature_type=1)
    - Safe/Gnosis wallets (signature_type=2)
    """

    def __init__(
        self,
        private_key: HexStr,
        signature_type: Literal[0, 1, 2] = 1,
        chain_id: Literal[137, 80002] = POLYGON,
        rpc_url: str = "https://tenderly.rpc.polygon.community",
    ):
        super().__init__(
            private_key, signature_type, chain_id=chain_id, rpc_url=rpc_url
        )

    def _execute(
        self,
        to: ChecksumAddress,
        data: str,
        operation_name: str,
        metadata: str | None = None,  # noqa: ARG002
    ) -> TransactionReceipt:
        """Execute transaction on-chain with gas."""
        base_transaction = self._build_base_transaction()

        match self.signature_type:
            case 0:
                txn_data = self._build_eoa_transaction(to, data, base_transaction)
            case 1:
                txn_data = self._build_proxy_transaction(to, data, base_transaction)
            case 2:
                txn_data = self._build_safe_transaction(to, data, base_transaction)
            case _:
                msg = f"Invalid signature_type: {self.signature_type}"
                raise ValueError(msg)

        return self._execute_transaction(txn_data, operation_name)

    def _build_base_transaction(self) -> TxParams:
        """Build base transaction parameters."""
        nonce = self.w3.eth.get_transaction_count(self.account.address)
        current_gas_price: int = self.w3.eth.gas_price
        adjusted_gas_price = Wei(int(current_gas_price * 1.05))

        return {
            "nonce": nonce,
            "gasPrice": adjusted_gas_price,
            "gas": 1000000,
            "from": self.account.address,
            "chainId": self.chain_id,
        }

    def _build_eoa_transaction(
        self, to: ChecksumAddress, data: str, base_transaction: TxParams
    ) -> TxParams:
        """Build transaction for EOA wallet."""
        estimation_txn: TxParams = {
            "from": self.address,
            "to": to,
            "data": HexStr(data),
        }

        estimated = self.w3.eth.estimate_gas(estimation_txn)
        base_transaction["gas"] = int(estimated * 1.05)
        base_transaction["to"] = to
        base_transaction["data"] = HexStr(data)

        return base_transaction

    def _build_proxy_transaction(
        self, to: ChecksumAddress, data: str, base_transaction: TxParams
    ) -> TxParams:
        """Build transaction for Poly proxy wallet."""
        proxy_txn = {
            "typeCode": 1,
            "to": to,
            "value": 0,
            "data": data,
        }

        estimation_txn: TxParams = {
            "from": self.address,
            "to": to,
            "data": HexStr(data),
        }
        estimated = self.w3.eth.estimate_gas(estimation_txn)
        base_transaction["gas"] = int(estimated * 1.05) + 100000

        txn_data = self.proxy_factory.functions.proxy([proxy_txn]).build_transaction(
            transaction=base_transaction
        )
        return txn_data

    def _build_safe_transaction(
        self, to: ChecksumAddress, data: str, base_transaction: TxParams
    ) -> TxParams:
        """Build transaction for Safe wallet."""
        safe_nonce = self.safe.functions.nonce().call()
        safe_txn: SafeTxn = {
            "to": to,
            "data": data,
            "operation": 0,
            "value": 0,
        }
        packed_sig = get_packed_signature(
            sign_safe_transaction(
                self.account,
                self.safe,
                safe_txn,
                safe_nonce,
            )
        )

        estimation_txn: TxParams = {
            "from": self.address,
            "to": to,
            "data": HexStr(data),
        }
        estimated = self.w3.eth.estimate_gas(estimation_txn)
        base_transaction["gas"] = int(estimated * 1.05) + 100000

        txn_data = self.safe.functions.execTransaction(
            safe_txn["to"],
            safe_txn["value"],
            safe_txn["data"],
            safe_txn.get("operation", 0),
            0,
            0,
            0,
            ADDRESS_ZERO,
            ADDRESS_ZERO,
            packed_sig,
        ).build_transaction(transaction=base_transaction)

        return txn_data

    def _execute_transaction(
        self, txn_data: TxParams, operation_name: str
    ) -> TransactionReceipt:
        """Execute transaction and wait for receipt."""
        signed_txn = self.account.sign_transaction(txn_data)
        tx_hash = self.w3.eth.send_raw_transaction(signed_txn.raw_transaction)
        tx_hash_hex = tx_hash.hex()

        print(f"Txn hash: 0x{tx_hash_hex}")

        receipt_dict = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        receipt = TransactionReceipt.model_validate(receipt_dict)

        print(
            f"{operation_name} succeeded"
            if receipt.status == 1
            else f"{operation_name} failed"
        )
        print(
            f"Paid {round((receipt.gas_used * receipt.effective_gas_price) / 10**18, 3)} POL for gas"
        )

        return receipt

    def set_collateral_approval(self, spender: ChecksumAddress) -> TransactionReceipt:
        """Set approval for spender on USDC."""
        to = self.usdc_address
        data = self._encode_usdc_approve(address=spender)
        return self._execute(to, data, "Collateral Approval")

    def set_conditional_tokens_approval(
        self, spender: ChecksumAddress
    ) -> TransactionReceipt:
        """Set approval for spender on conditional tokens."""
        to = self.conditional_tokens_address
        data = self._encode_condition_tokens_approve(address=spender)
        return self._execute(to, data, "Conditional Tokens Approval")

    def set_all_approvals(self) -> list[TransactionReceipt]:
        """Set all necessary approvals."""
        receipts = []
        print("Approving ConditionalTokens as spender on USDC")
        receipts.append(
            self.set_collateral_approval(spender=self.conditional_tokens_address)
        )
        print("Approving CTFExchange as spender on USDC")
        receipts.append(self.set_collateral_approval(spender=self.exchange_address))
        print("Approving NegRiskCtfExchange as spender on USDC")
        receipts.append(
            self.set_collateral_approval(spender=self.neg_risk_exchange_address)
        )
        print("Approving NegRiskAdapter as spender on USDC")
        receipts.append(
            self.set_collateral_approval(spender=self.neg_risk_adapter_address)
        )
        print("Approving CTFExchange as spender on ConditionalTokens")
        receipts.append(
            self.set_conditional_tokens_approval(spender=self.exchange_address)
        )
        print("Approving NegRiskCtfExchange as spender on ConditionalTokens")
        receipts.append(
            self.set_conditional_tokens_approval(spender=self.neg_risk_exchange_address)
        )
        print("Approving NegRiskAdapter as spender on ConditionalTokens")
        receipts.append(
            self.set_conditional_tokens_approval(spender=self.neg_risk_adapter_address)
        )
        print("All approvals set!")
        return receipts

    def transfer_usdc(self, recipient: EthAddress, amount: float) -> TransactionReceipt:
        """Transfer USDC to recipient."""
        balance = self.get_usdc_balance(address=self.address)
        if balance < amount:
            msg = f"Insufficient USDC balance: {balance} < {amount}"
            raise ValueError(msg)

        amount_int = int(amount * 1e6)
        to = self.usdc_address
        data = self._encode_transfer_usdc(
            self.w3.to_checksum_address(recipient), amount_int
        )
        return self._execute(to, data, "USDC Transfer")

    def transfer_token(
        self, token_id: str, recipient: EthAddress, amount: float
    ) -> TransactionReceipt:
        """Transfer conditional token to recipient."""
        balance = self.get_token_balance(token_id=token_id, address=self.address)
        if balance < amount:
            msg = f"Insufficient token balance: {balance} < {amount}"
            raise ValueError(msg)

        amount_int = int(amount * 1e6)
        to = self.conditional_tokens_address
        data = self._encode_transfer_token(
            token_id, self.w3.to_checksum_address(recipient), amount_int
        )
        return self._execute(to, data, "Token Transfer")

    def deploy_safe(self) -> TransactionReceipt:
        """Deploy a Safe wallet."""
        safe_address = self.get_safe_proxy_address()
        if self.w3.eth.get_code(self.w3.to_checksum_address(safe_address)) != b"":
            msg = f"Safe already deployed at {safe_address}"
            raise SafeAlreadyDeployedError(msg)

        sig = create_safe_create_signature(account=self.account, chain_id=POLYGON)
        split_sig = split_signature(sig)

        base_transaction = self._build_base_transaction()
        txn_data = self.safe_proxy_factory.functions.createProxy(
            ADDRESS_ZERO,
            0,
            ADDRESS_ZERO,
            (split_sig["v"], split_sig["r"], split_sig["s"]),
        ).build_transaction(transaction=base_transaction)

        return self._execute_transaction(txn_data, "Gnosis Safe Deployment")


class PolymarketGaslessWeb3Client(BaseWeb3Client):
    """Polymarket Web3 client for gasless transactions via relay."""

    def __init__(
        self,
        private_key: HexStr,
        signature_type: Literal[1, 2] = 1,
        *,
        relayer_api_key: str,
        chain_id: Literal[137, 80002] = POLYGON,
        rpc_url: str = "https://tenderly.rpc.polygon.community",
    ):
        if signature_type not in {1, 2}:
            msg = "PolymarketGaslessWeb3Client only supports signature_type=1 (Poly proxy wallets) and signature_type=2 (Safe wallets)."
            raise ValueError(msg)

        super().__init__(
            private_key, signature_type, chain_id=chain_id, rpc_url=rpc_url
        )

        # Setup for gasless transactions
        self.signer = Signer(private_key=private_key, chain_id=chain_id)
        self.relay_url = "https://relayer-v2.polymarket.com"
        self.relay_hub = "0xD216153c06E857cD7f72665E0aF1d7D82172F494"
        self.relay_address = "0x7db63fe6d62eb73fb01f8009416f4c2bb4fbda6a"
        self.relayer_api_key = relayer_api_key

    def _execute(
        self,
        to: ChecksumAddress,
        data: str,
        operation_name: str,
        metadata: str | None = None,
    ) -> TransactionReceipt:
        """Execute transaction via gasless relay."""
        match self.signature_type:
            case 1:
                body = self._build_proxy_relay_transaction(to, data, metadata or "")
            case 2:
                body = self._build_safe_relay_transaction(to, data, metadata or "")
            case _:
                msg = f"Invalid signature_type: {self.signature_type}"
                raise ValueError(msg)

        headers = create_relayer_headers(
            self.relayer_api_key, self.get_base_address()
        )

        url = f"{self.relay_url}/submit"
        response = self.client.post(
            url, headers=headers, content=dumps(body).encode("utf-8")
        )
        response.raise_for_status()

        gasless_response = response.json()

        print(
            f"Gasless txn submitted: {gasless_response.get('transactionHash', 'N/A')}"
        )
        print(f"Transaction ID: {gasless_response.get('transactionID', 'N/A')}")
        print(f"State: {gasless_response.get('state', 'N/A')}")

        # Wait for confirmation and return receipt
        tx_hash = gasless_response.get("transactionHash")
        if tx_hash:
            receipt_dict = self.w3.eth.wait_for_transaction_receipt(tx_hash)
            receipt = TransactionReceipt.model_validate(receipt_dict)

            print(
                f"{operation_name} succeeded"
                if receipt.status == 1
                else f"{operation_name} failed"
            )

            return receipt
        msg = f"No transaction hash in response: {gasless_response}"
        raise ValueError(msg)

    def _get_relay_nonce(self, wallet_type: Literal["PROXY", "SAFE"]) -> int:
        """Get nonce from relay for Safe wallet."""
        url = f"{self.relay_url}/nonce"
        params = {
            "address": self.get_base_address(),
            "type": wallet_type,
        }
        response = self.client.get(url, params=params)
        response.raise_for_status()
        return int(response.json()["nonce"])

    def _build_proxy_relay_transaction(
        self, to: ChecksumAddress, data: str, metadata: str
    ) -> dict[str, object]:
        """Build Proxy relay transaction body."""
        proxy_nonce = self._get_relay_nonce(wallet_type="PROXY")
        gas_price = "0"
        relayer_fee = "0"

        proxy_txn: dict[str, object] = {
            "typeCode": 1,
            "to": to,
            "value": 0,
            "data": data,
        }

        encoded_txn = self._encode_proxy(proxy_txn)

        try:
            estimation_txn: TxParams = {
                "from": self.get_base_address(),
                "to": self.proxy_factory_address,
                "data": HexStr(encoded_txn),
            }
            estimated_gas = self.w3.eth.estimate_gas(estimation_txn)
            gas_limit = str(int(estimated_gas * 1.3 + 100000))
        except TimeExhausted as e:
            print(
                f"Timeout during gas estimation for proxy transaction, using default: {e}"
            )
            gas_limit = str(10_000_000)

        struct = create_proxy_struct(
            from_address=self.get_base_address(),
            to=self.proxy_factory_address,
            data=encoded_txn,
            tx_fee=relayer_fee,
            gas_price=gas_price,
            gas_limit=gas_limit,
            nonce=str(proxy_nonce),
            relay_hub_address=self.relay_hub,
            relay_address=self.relay_address,
        )

        struct_hash = "0x" + self.w3.keccak(struct).hex()

        signature = self.account.sign_message(
            encode_defunct(hexstr=struct_hash)
        ).signature.hex()

        return {
            "data": encoded_txn,
            "from": self.get_base_address(),
            "metadata": metadata,
            "nonce": str(proxy_nonce),
            "proxyWallet": self.get_poly_proxy_address(),
            "signature": "0x" + signature,
            "signatureParams": {
                "gasPrice": gas_price,
                "gasLimit": gas_limit,
                "relayerFee": relayer_fee,
                "relayHub": self.relay_hub,
                "relay": self.relay_address,
            },
            "to": self.proxy_factory_address,
            "type": "PROXY",
        }

    def _build_safe_relay_transaction(
        self, to: ChecksumAddress, data: str, metadata: str
    ) -> dict[str, object]:
        """Build Safe relay transaction body."""
        safe_nonce = self._get_relay_nonce(wallet_type="SAFE")

        safe_txn: SafeTxn = {
            "to": to,
            "data": data,
            "operation": 0,
            "value": 0,
        }

        signature = sign_safe_transaction(
            self.account,
            self.safe,
            safe_txn,
            safe_nonce,
        ).signature.hex()

        match signature[-2:]:
            case "00" | "1b":
                signature = signature[:-2] + "1f"
            case "01" | "1c":
                signature = signature[:-2] + "20"

        return {
            "data": safe_txn["data"],
            "from": self.get_base_address(),
            "metadata": metadata,
            "nonce": str(safe_nonce),
            "proxyWallet": self.get_safe_proxy_address(),
            "signature": "0x" + signature,
            "signatureParams": {
                "baseGas": "0",
                "gasPrice": "0",
                "gasToken": ADDRESS_ZERO,
                "operation": "0",
                "refundReceiver": ADDRESS_ZERO,
                "safeTxnGas": "0",
            },
            "to": to,
            "type": "SAFE",
        }



"""Testing helpers for live contract checks and other SDK test utilities."""

from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .contract_assertions import assert_api_contract, fail_contract, fetch_json

__all__ = ["assert_api_contract", "fail_contract", "fetch_json"]

_EXPORT_MAP = {
    "assert_api_contract": ".contract_assertions",
    "fail_contract": ".contract_assertions",
    "fetch_json": ".contract_assertions",
}


def __getattr__(name: str) -> Any:
    module_name = _EXPORT_MAP.get(name)
    if module_name is None:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    module = import_module(module_name, __name__)
    value = getattr(module, name)
    globals()[name] = value
    return value



from datetime import UTC, datetime
from typing import Literal, Optional, cast

from pydantic import BaseModel, Field, field_validator, model_validator

from .common import EmptyString, EthAddress, Keccak256

# Type aliases
type ActivityType = Literal[
    "TRADE", "SPLIT", "MERGE", "REDEEM", "REWARD", "CONVERSION", "MAKER_REBATE", "YIELD"
]


class AccountingSnapshotCSVs(BaseModel):
    """Parsed contents of the accounting snapshot ZIP."""

    positions_csv: str
    equity_csv: str


class GQLPosition(BaseModel):
    user: EthAddress
    token_id: str
    complementary_token_id: str
    condition_id: Keccak256
    outcome_index: int
    balance: float

    @model_validator(mode="before")
    def _flatten(cls, values: dict[str, object]) -> dict[str, object]:
        asset = values.get("asset")
        if isinstance(asset, dict):
            if "id" in asset:
                values.setdefault("token_id", asset["id"])
            if "complement" in asset:
                values.setdefault("complementary_token_id", asset["complement"])
            condition = asset.get("condition")
            if isinstance(condition, dict) and "id" in condition:
                values.setdefault("condition_id", condition["id"])
            if "outcomeIndex" in asset:
                values.setdefault("outcome_index", asset["outcomeIndex"])
            values.pop("asset", None)
        return values

    @field_validator("balance", mode="before")
    @classmethod
    def _parse_balance(cls, value: str | float) -> float:
        if isinstance(value, str):
            value = int(value)
        return value / 10**6


class Position(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Asset information
    token_id: str = Field(alias="asset")
    complementary_token_id: str = Field(alias="oppositeAsset")
    condition_id: Keccak256 = Field(alias="conditionId")
    event_id: int | None = Field(None, alias="eventId")
    outcome: str
    complementary_outcome: str = Field(alias="oppositeOutcome")
    outcome_index: int = Field(alias="outcomeIndex")

    # Position details
    size: float
    avg_price: float = Field(alias="avgPrice")
    current_price: float = Field(alias="curPrice")
    redeemable: bool
    mergeable: bool = False

    # Financial metrics
    initial_value: float = Field(alias="initialValue")
    current_value: float = Field(alias="currentValue")
    cash_pnl: float = Field(alias="cashPnl")
    percent_pnl: float = Field(alias="percentPnl")
    total_bought: float = Field(alias="totalBought")
    realized_pnl: float = Field(alias="realizedPnl")
    percent_realized_pnl: float = Field(alias="percentRealizedPnl")

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    end_date: datetime = Field(alias="endDate")
    negative_risk: bool = Field(alias="negativeRisk")

    @field_validator("end_date", mode="before")
    def handle_empty_end_date(cls, v: datetime | Literal[""]) -> datetime:
        if v == "":
            return datetime(2099, 12, 31, tzinfo=UTC)
        if isinstance(v, datetime):
            return v
        return cast("datetime", v)


class ClosedPosition(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Asset information
    token_id: str = Field(alias="asset")
    complementary_token_id: str = Field(alias="oppositeAsset")
    condition_id: Keccak256 = Field(alias="conditionId")
    outcome: str
    complementary_outcome: str = Field(alias="oppositeOutcome")
    outcome_index: int = Field(alias="outcomeIndex")

    # Position details
    avg_price: float = Field(alias="avgPrice")
    current_price: float = Field(alias="curPrice")
    timestamp: datetime | None = None

    # Financial metrics
    total_bought: float = Field(alias="totalBought")
    realized_pnl: float = Field(alias="realizedPnl")

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    end_date: datetime = Field(alias="endDate")

    @field_validator("end_date", mode="before")
    def handle_empty_end_date(cls, v: datetime | Literal[""]) -> datetime:
        if v == "":
            return datetime(2099, 12, 31, tzinfo=UTC)
        if isinstance(v, datetime):
            return v
        return cast("datetime", v)


class Trade(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Trade details
    side: Literal["BUY", "SELL"]
    token_id: str = Field(alias="asset")
    condition_id: Keccak256 = Field(alias="conditionId")
    size: float
    price: float
    timestamp: datetime

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    outcome: str
    outcome_index: int = Field(alias="outcomeIndex")

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")

    # Transaction information
    transaction_hash: Keccak256 = Field(alias="transactionHash")


class Activity(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Activity details
    timestamp: datetime
    condition_id: Keccak256 | EmptyString = Field(alias="conditionId")
    type: ActivityType
    size: float
    usdc_size: float = Field(alias="usdcSize")
    price: float
    asset: str
    side: str | None
    outcome_index: int = Field(alias="outcomeIndex")

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    outcome: str

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")

    # Transaction information
    transaction_hash: Keccak256 = Field(alias="transactionHash")


class Holder(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Holder details
    token_id: str = Field(alias="asset")
    amount: float
    outcome_index: int = Field(alias="outcomeIndex")

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")
    display_username_public: bool = Field(alias="displayUsernamePublic")
    verified: bool | None = None


class HolderResponse(BaseModel):
    # Asset information
    token_id: str = Field(alias="token")

    # Holders list
    holders: list[Holder]


class ValueResponse(BaseModel):
    # User identification
    user: EthAddress

    # Value information
    value: float


class User(BaseModel):
    proxy_wallet: EthAddress = Field(alias="proxyWallet")
    name: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")


class UserMetric(User):
    amount: float
    pseudonym: str


class UserRank(User):
    amount: float
    rank: int


class UserID(BaseModel):
    id: str
    creator: bool
    mod: bool
    community_mod: Optional[bool] = Field(None, alias="communityMod")


class UserProfile(BaseModel):
    created_at: datetime = Field(alias="createdAt")
    proxy_wallet: EthAddress = Field(alias="proxyWallet")
    profile_image: Optional[str] = Field(None, alias="profileImage")
    display_username_public: bool = Field(alias="displayUsernamePublic")
    bio: Optional[str] = None
    pseudonym: str
    name: Optional[str] = None
    users: Optional[list[UserID]] = None
    x_username: Optional[str] = Field(None, alias="xUsername")
    verified_badge: bool = Field(alias="verifiedBadge")


class LeaderboardUser(BaseModel):
    rank: int
    proxy_wallet: EthAddress = Field(alias="proxyWallet")
    username: str = Field(alias="userName")
    x_username: str = Field(alias="xUsername")
    verified_badge: bool = Field(alias="verifiedBadge")
    vol: float
    pnl: float
    profile_image: str = Field(alias="profileImage")


class BuilderLeaderboardUser(BaseModel):
    date: Optional[datetime] = Field(None, alias="dt")  # period end date
    rank: int
    builder: str
    volume: float
    active_users: int = Field(alias="activeUsers")
    verified: float
    builder_logo: str = Field(alias="builderLogo")


class MarketValue(BaseModel):
    condition_id: Keccak256 = Field(alias="market")
    value: float


class EventLiveVolume(BaseModel):
    total: Optional[float]
    markets: Optional[list[MarketValue]]



from __future__ import annotations

import inspect
import json
import keyword
import os
import re
from dataclasses import dataclass
from pathlib import Path
from types import NoneType, UnionType
from typing import Annotated, Any, NoReturn, Union, get_args, get_origin

import httpx
import pytest
from pydantic import BaseModel, RootModel, TypeAdapter, ValidationError
from pydantic.aliases import AliasChoices, AliasPath

SNAPSHOT_DIR = Path(__file__).resolve().parents[3] / "tests" / "prod_read" / "snapshots"
AUTO_EXPAND_CONTRACT = os.getenv("AUTO_EXPAND_CONTRACT") == "1"
UPDATE_SNAPSHOTS = os.getenv("UPDATE_SCHEMA_SNAPSHOTS") == "1" or AUTO_EXPAND_CONTRACT
EXPANSION_REPORT_PATH = (
    Path(__file__).resolve().parents[3]
    / "tests"
    / "prod_read"
    / "contract_expansion_report.json"
)

_REPORT_ROWS: list[dict[str, str]] = []
_PATCHED_FIELDS: set[tuple[str, str, str]] = set()


@dataclass(frozen=True)
class UnknownFieldObservation:
    owner_model: type[BaseModel]
    path: str
    alias: str
    sample_value: Any


def fetch_json(
    name: str,
    client: httpx.Client,
    url: str,
    *,
    params: dict[str, Any] | None = None,
) -> Any:
    try:
        response = client.get(url, params=params)
        response.raise_for_status()
    except httpx.HTTPStatusError as exc:
        fail_contract(
            "endpoint unavailable",
            (
                f"{name} returned HTTP {exc.response.status_code}.\n"
                f"URL: {exc.request.url}\n"
                "This looks like endpoint instability, auth/rate limiting, or an upstream outage."
            ),
        )
    except httpx.RequestError as exc:
        fail_contract(
            "endpoint unavailable",
            (
                f"{name} request failed before a valid response was received.\n"
                f"URL: {exc.request.url}\n"
                f"Error: {exc!r}"
            ),
        )
    else:
        return response.json()


def assert_api_contract(name: str, annotation: Any, payload: Any) -> Any:
    try:
        validated = TypeAdapter(annotation).validate_python(payload)
    except ValidationError as exc:
        fail_contract("schema mismatch", _validation_message(name, payload, exc))

    observed_shape = sorted(_collect_shape(payload))
    snapshot_path = SNAPSHOT_DIR / f"{_snapshot_slug(name)}.json"
    known_shape = _read_known_shape(snapshot_path)
    added_paths = sorted(set(observed_shape) - set(known_shape))

    unknown_observations = _collect_unknown_field_observations(annotation, payload)

    if AUTO_EXPAND_CONTRACT:
        _merge_shape_snapshot(snapshot_path, name, observed_shape)
        if unknown_observations:
            _apply_model_patches(name, unknown_observations)
        return validated

    if UPDATE_SNAPSHOTS:
        _merge_shape_snapshot(snapshot_path, name, observed_shape)
        return validated

    if added_paths or unknown_observations:
        fail_contract(
            "contract expanded",
            _contract_expanded_message(
                name,
                snapshot_path,
                added_paths,
                unknown_observations,
            ),
        )

    return validated


def fail_contract(category: str, message: str) -> NoReturn:
    pytest.fail(f"[{category}] {message}")


def _validation_message(name: str, payload: Any, exc: ValidationError) -> str:
    first_error = exc.errors()[0]
    return (
        f"{name} schema drift detected.\n"
        f"Validation path: {first_error['loc']}\n"
        f"Validation message: {first_error['msg']}\n"
        f"Payload type: {type(payload).__name__}\n"
        "Run this test locally to inspect the full response body."
    )


def _contract_expanded_message(
    name: str,
    snapshot_path: Path,
    added_paths: list[str],
    unknown_observations: list[UnknownFieldObservation],
) -> str:
    message = (
        f"{name} expanded beyond the currently modeled contract.\n"
        f"Snapshot: {snapshot_path}\n"
    )
    if added_paths:
        message += f"New paths/types:\n{_format_items(added_paths)}\n"
    if unknown_observations:
        message += "New unmapped model fields:\n"
        message += _format_items(
            [
                f"{obs.path} -> {obs.owner_model.__module__}.{obs.owner_model.__name__}"
                for obs in unknown_observations
            ]
        )
        message += "\n"
    message += (
        "Run the auto-expansion workflow to update snapshots and generate a PR "
        "for new model fields."
    )
    return message


def _collect_shape(payload: Any, path: str = "$") -> set[str]:
    shape = {f"{path}:{_json_type_name(payload)}"}

    if isinstance(payload, dict):
        for key, value in sorted(payload.items()):
            shape |= _collect_shape(value, f"{path}.{key}")
        return shape

    if isinstance(payload, list):
        for item in payload:
            shape |= _collect_shape(item, f"{path}[]")
        return shape

    return shape


def _read_known_shape(snapshot_path: Path) -> list[str]:
    if not snapshot_path.exists():
        return []
    payload = json.loads(snapshot_path.read_text())
    shape = payload.get("shape", [])
    if isinstance(shape, list):
        return [item for item in shape if isinstance(item, str)]
    return []


def _merge_shape_snapshot(snapshot_path: Path, name: str, observed_shape: list[str]) -> None:
    known_shape = _read_known_shape(snapshot_path)
    merged = sorted(set(known_shape) | set(observed_shape))
    snapshot_path.parent.mkdir(parents=True, exist_ok=True)
    snapshot_path.write_text(json.dumps({"name": name, "shape": merged}, indent=2) + "\n")


def _collect_unknown_field_observations(
    annotation: Any,
    payload: Any,
    path: str = "$",
) -> list[UnknownFieldObservation]:
    annotation = _unwrap_annotation(annotation)

    if payload is None or annotation in {Any, object}:
        return []

    if _is_root_model(annotation):
        return _collect_unknown_field_observations(
            annotation.model_fields["root"].annotation,
            payload,
            path,
        )

    if _is_model(annotation):
        if not isinstance(payload, dict):
            return []

        known_fields: dict[str, Any] = {}
        for field_name, field in annotation.model_fields.items():
            for candidate in _field_input_names(field_name, field):
                known_fields[candidate] = field

        observations = [
            UnknownFieldObservation(
                owner_model=annotation,
                path=f"{path}.{key}",
                alias=key,
                sample_value=payload[key],
            )
            for key in sorted(payload)
            if key not in known_fields
        ]

        for key, value in payload.items():
            field = known_fields.get(key)
            if field is None:
                continue
            observations.extend(
                _collect_unknown_field_observations(
                    field.annotation,
                    value,
                    f"{path}.{key}",
                )
            )
        return observations

    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin in {list, tuple, set, frozenset} and args and isinstance(payload, list):
        item_type = args[0]
        nested_observations: list[UnknownFieldObservation] = []
        for item in payload:
            nested_observations.extend(
                _collect_unknown_field_observations(item_type, item, f"{path}[]")
            )
        return nested_observations

    if origin is dict:
        return []

    if origin in {UnionType, Union}:
        branch_results = [
            _collect_unknown_field_observations(branch, payload, path)
            for branch in args
            if branch is not NoneType
        ]
        if not branch_results:
            return []
        return min(branch_results, key=len)

    return []


def _apply_model_patches(
    endpoint_name: str,
    observations: list[UnknownFieldObservation],
) -> None:
    for observation in observations:
        owner_model = observation.owner_model
        file_path = inspect.getsourcefile(owner_model)
        if file_path is None:
            continue

        field_name = _normalize_field_name(observation.alias)
        patch_key = (file_path, owner_model.__name__, field_name)
        if patch_key in _PATCHED_FIELDS:
            continue

        source_path = Path(file_path)
        inserted = _insert_field_into_model(
            source_path,
            owner_model.__name__,
            field_name,
            observation.alias,
            observation.sample_value,
        )
        if not inserted:
            continue

        _PATCHED_FIELDS.add(patch_key)
        _REPORT_ROWS.append(
            {
                "endpoint": endpoint_name,
                "model": f"{owner_model.__module__}.{owner_model.__name__}",
                "path": observation.path,
                "alias": observation.alias,
                "field_name": field_name,
                "type": _infer_field_type(observation.sample_value),
                "file": str(source_path),
            }
        )

    if _REPORT_ROWS:
        EXPANSION_REPORT_PATH.parent.mkdir(parents=True, exist_ok=True)
        EXPANSION_REPORT_PATH.write_text(json.dumps(_REPORT_ROWS, indent=2) + "\n")


def _insert_field_into_model(
    source_path: Path,
    class_name: str,
    field_name: str,
    alias: str,
    sample_value: Any,
) -> bool:
    lines = source_path.read_text().splitlines()
    class_index = _find_class_index(lines, class_name)
    if class_index is None:
        return False

    class_end = _find_class_end(lines, class_index)
    if _class_has_field(lines[class_index:class_end], field_name):
        return False

    insert_at = _find_insert_index(lines, class_index, class_end)
    rendered_field = _render_field_line(field_name, alias, sample_value)
    lines.insert(insert_at, rendered_field)
    source_path.write_text("\n".join(lines) + "\n")
    return True


def _find_class_index(lines: list[str], class_name: str) -> int | None:
    class_pattern = re.compile(rf"^class {re.escape(class_name)}\b")
    for index, line in enumerate(lines):
        if class_pattern.match(line):
            return index
    return None


def _find_class_end(lines: list[str], class_index: int) -> int:
    for index in range(class_index + 1, len(lines)):
        stripped = lines[index].strip()
        if not stripped:
            continue
        indent = len(lines[index]) - len(lines[index].lstrip())
        if indent == 0:
            return index
    return len(lines)


def _find_insert_index(lines: list[str], class_index: int, class_end: int) -> int:
    for index in range(class_index + 1, class_end):
        stripped = lines[index].lstrip()
        if stripped.startswith(("@", "def ")):
            return index
    return class_end


def _class_has_field(class_lines: list[str], field_name: str) -> bool:
    field_pattern = re.compile(rf"^\s+{re.escape(field_name)}:")
    return any(field_pattern.match(line) for line in class_lines)


def _render_field_line(field_name: str, alias: str, sample_value: Any) -> str:
    field_type = _infer_field_type(sample_value)
    if alias == field_name:
        return f"    {field_name}: {field_type} = None"
    return f'    {field_name}: {field_type} = Field(None, alias="{alias}")'


def _infer_field_type(value: Any) -> str:
    if isinstance(value, bool):
        return "bool | None"
    if isinstance(value, int):
        return "int | None"
    if isinstance(value, float):
        return "float | None"
    if isinstance(value, str):
        return "str | None"
    if isinstance(value, list):
        return "list[object] | None"
    if isinstance(value, dict):
        return "dict[str, object] | None"
    return "object | None"


def _normalize_field_name(alias: str) -> str:
    field_name = re.sub(r"(?<!^)(?=[A-Z])", "_", alias).lower()
    field_name = re.sub(r"[^0-9a-zA-Z_]", "_", field_name)
    field_name = re.sub(r"_+", "_", field_name).strip("_") or "field"
    if field_name[0].isdigit():
        field_name = f"field_{field_name}"
    if keyword.iskeyword(field_name):
        field_name = f"{field_name}_field"
    return field_name


def _unwrap_annotation(annotation: Any) -> Any:
    while True:
        origin = get_origin(annotation)
        if origin is Annotated:
            annotation = get_args(annotation)[0]
            continue
        return annotation


def _field_input_names(field_name: str, field: Any) -> set[str]:
    names = {field_name}
    if field.alias:
        names.add(field.alias)

    validation_alias = field.validation_alias
    if isinstance(validation_alias, str):
        names.add(validation_alias)
    elif isinstance(validation_alias, AliasChoices):
        for choice in validation_alias.choices:
            if isinstance(choice, str):
                names.add(choice)
    elif (
        isinstance(validation_alias, AliasPath)
        and validation_alias.path
        and isinstance(validation_alias.path[0], str)
    ):
        names.add(validation_alias.path[0])

    return names


def _json_type_name(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, bool):
        return "bool"
    if isinstance(value, int | float):
        return "number"
    if isinstance(value, str):
        return "str"
    if isinstance(value, dict):
        return "object"
    if isinstance(value, list):
        return "array"
    return type(value).__name__


def _is_model(annotation: Any) -> bool:
    return isinstance(annotation, type) and issubclass(annotation, BaseModel)


def _is_root_model(annotation: Any) -> bool:
    return isinstance(annotation, type) and issubclass(annotation, RootModel)


def _snapshot_slug(name: str) -> str:
    return name.lower().replace("/", "_").replace("{", "").replace("}", "").replace(" ", "_")


def _format_items(items: list[str]) -> str:
    return "\n".join(f"- {item}" for item in items[:40]) + (
        "\n- ..." if len(items) > 40 else ""
    )



class InvalidPriceError(Exception):
    pass


class InvalidTickSizeError(Exception):
    pass


class InvalidFeeRateError(Exception):
    pass


class LiquidityError(Exception):
    pass


class MissingOrderbookError(Exception):
    pass


class SafeAlreadyDeployedError(Exception):
    """Raised when attempting to deploy a Safe that has already been deployed."""


class BuilderRateLimitError(Exception):
    """Shared builder credentials have hit their rate limit."""

